/* This program calculates the eigenvalues of the matrix system:

                ([K] - lambda*[M])*[z] = [0]

        [K] = nXn  stiffness matrix
        [M] = nXn  mass matrix
        [z] = nX1  eigenvector
        lambda = eigenvalue

   using the Lanczos method. 

                        Updated 2/1/00 

    SLFFEA source file
    Version:  1.1
    Copyright (C) 1999  San Le

    The source code contained in this file is released under the
    terms of the GNU Library General Public License.

*/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define SMALL               1.0e-12
#define SMALLER             1.0e-20

int dotX(double *, double *, double *, int);
int matX(double *, double *, double *, int, int, int);
int matXT(double *, double *, double *, int, int, int);
int solve(double *, double *, int *, int );
int QR_check( double *, double *, double *, int );

typedef struct {
	double val;
	int index;
} EIGEN;

int compare_eigen ( const void *x,  const void *y)
{
	EIGEN *a = (EIGEN *) x;
	EIGEN *b = (EIGEN *) y;
	if ( a[0].val < b[0].val ) return -1;
	if ( a[0].val == b[0].val ) return 0;
	if ( a[0].val > b[0].val ) return 1;
}

int lanczos(double *K, double *M, int *idiag, double *ritz,
	int n, int num_eigen)
{
	int i,i2,j,jsave,k,check;
        int sofmf, sofeigen, ptr_inc,sof;
        int *index;
	double *mem_double;
	double *r, *rbar, *rhat, *p, *pbar, *qm1, *Mq, *vv,
		*q, *T_evector, *T_diag, *T_upper, *temp;
	double lamda, dum, dum2, fdum, alpha, beta, err, d, dn;
	EIGEN *eigen, *mem_eigen;


	sof = sizeof(double);

        index=(int *)malloc(n*sizeof(int));
        if(!index) printf( "error allocating memory for index\n");

        sofmf= num_eigen*num_eigen + 2*n*num_eigen + 5*n +
		num_eigen + num_eigen - 1;
        mem_double=(double *)calloc(sofmf,sizeof(double));
        if(!mem_double )
        {
                printf( "failed to allocate memory for double\n ");
                exit(1);
        }
                                           ptr_inc=0;
        T_evector=(mem_double+ptr_inc);    ptr_inc += num_eigen*num_eigen;
        q=(mem_double+ptr_inc);            ptr_inc += n*num_eigen;
        Mq=(mem_double+ptr_inc);           ptr_inc += n*num_eigen;
        p=(mem_double+ptr_inc);            ptr_inc += n;
        qm1=(mem_double+ptr_inc);          ptr_inc += n;
        r=(mem_double+ptr_inc);            ptr_inc += n;
        vv=(mem_double+ptr_inc);           ptr_inc += n;
        temp=(mem_double+ptr_inc);         ptr_inc += n;
        T_diag=(mem_double+ptr_inc);       ptr_inc += num_eigen;
        T_upper=(mem_double+ptr_inc);      ptr_inc += num_eigen-1;

/* Allocate Memory for the eigens */
	sofeigen= num_eigen;
	mem_eigen=(EIGEN *)calloc(sofeigen,sizeof(EIGEN));

	if(!mem_eigen )
	{
		printf( "failed to allocate memory for eigen\n ");
		exit(1);
	}
                                          ptr_inc=0;
	eigen=(mem_eigen+ptr_inc);        ptr_inc += num_eigen;

/* Initialize variables */

	for( i = 0; i < num_eigen; ++i )
	{
		eigen[i].index = i;
	}

	dn = (double) n;

	fdum = 0.0;
        for( i = 0; i < n; ++i ) 
	{
		*(qm1 + i) = 0.0;
		*(r + i) = sin(3.14159*fdum);
		fdum += 3.14159/dn;
	}
	/* *(r + 0) = 1.0;*/

/* Set p = M*r */
#if 0
	check = matX(p, M, r, n, 1, n);
        if(!check) printf( " Problems with matX \n");
#endif
        for( i = 0; i < n; ++i ) 
	{
		*(p + i) = *(M + i)*(*(r + i));
	}

	check = dotX( &beta, r, p, n);
        if(!check) printf( " Problems with dotX \n");
	beta = sqrt(beta);

        for( i = 0; i < n; ++i ) 
	{
		*(q + i ) = *(r+i)/beta;

/* p = M*q = M*r/beta */

		*(p + i ) /= beta;
	}

        for( i = 0; i < n; ++i ) 
	{
		*(Mq + i) = *(p + i);
		*(r + i) = *(p + i);
	}

	err = 10000;
	dum = 0.0;

	err = 10000;
	dum = 0.0;
	k = 0;

/* Begin iteration */

        /*while( err > 1.0e-6 )*/
        for( i = 0; i < num_eigen; ++i )
        {

/* Calculate K(inverse)y */
            check = solve(K, r, idiag, n);
            if(!check) printf( " Problems with solve \n");

            for( j = 0; j < n; ++j )
            {
	    	*(r + j) -= *(qm1 + j)*beta;
	    }

	    check = dotX( &alpha, p, r, n);
            if(!check) printf( " Problems with dotX \n");

/* Cretate tridiagonal T matrix */

	    *(T_diag + i) = alpha;

	    if(i)
	    {
	    	*(T_upper + i - 1 ) = beta;
	    }

            for( j = 0; j < n; ++j )
            {
	    	*(r + j) -= *(q + n*i + j)*alpha;
	    }

#if 0
	    check = matX(p, M, r, n, 1, n);
            if(!check) printf( " Problems with matX \n");
#endif
            for( j = 0; j < n; ++j ) 
	    {
		*(p + j) = *(M + j)*(*(r + j));
	    }

	    check = dotX( &beta, p, r, n);
            if(!check) printf( " Problems with dotX \n");
	    beta = sqrt(beta);

	    memset(temp,0,n*sof);
	    if( i < num_eigen - 1)
	    {
            	for( j = 0; j < n; ++j )
            	{
	    		*(qm1 + j) = *(q + n*i + j);
			*(q + n*(i+1) + j) = *(r + j)/beta;
                }

/* Re-orthogonalize q using Gram-Schmidt */

                printf( "\n");
                for( k = 0; k < i+1; ++k )
                {
                    check = dotX( &fdum, (Mq + n*k), (q + n*(i+1)), n);
            	    if(!check) printf( " Problems with dotX \n");
                    /*printf( "q(%d)M*q(%d) = %14.6e \n", k, i+1, fdum);*/
		    if ( fabs(fdum) > SMALL )
		    {
                    	for( i2 = 0; i2 < n; ++i2 )
                    	{
                        	*(temp + i2) += fdum*(*(q + n*k + i2));
				/*printf( " temp i2 %14.6e \n", *(temp + i2), i2);*/
                    	}
                    }
                }
            	for( j = 0; j < n; ++j )
            	{
		    *(q + n*(i+1) + j) -= *(temp + j);
                }

/* Calculate fdum = q*M*q and normalize q with q = q/sqrt(fdum) */

#if 0
		check = matX( temp, M, (q + n*(i+1)), n, 1, n);
            	if(!check) printf( " Problems with matX \n");
#endif
            	for( j = 0; j < n; ++j ) 
	    	{
			*(temp + j) = *(M + j)*(*(q + n*(i+1) + j));
	    	}

                check = dotX( &fdum, temp, (q + n*(i+1)), n);
            	if(!check) printf( " Problems with dotX \n");
		fdum = sqrt(fdum);
		for( j = 0; j < n; ++j )
                {
			*(q + n*(i+1) + j) /= fdum;
			*(p + j) =  *(temp + j)/fdum;
	    		*(Mq + n*(i+1) + j) = *(p + j);
			*(r + j) = *(p + j);
	    	}
	    }
	}

/* Calculate eigenvalues */

        printf("This is the T matrix before QR decomposition \n");
        for( i = 0; i < num_eigen; ++i )
        {
            printf( "%3d  ", i);
            for( j = 0; j < i-1; ++j )
            {
                 printf( "%14.6e ",0.0);
	    }
	    if(i) printf( "%14.6e ",*(T_upper+ i-1));
            printf( "%14.6e ",*(T_diag+i));
	    if(i < num_eigen - 1)
            {
            	printf( "%14.6e ",*(T_upper+i));
	    }
            for( j = i+1; j < num_eigen-1; ++j )
            {
                 printf( "%14.6e ",0.0);
	    }
            printf( "\n");
        }
        printf( "\n");

	check = QR_check(T_evector, T_diag, T_upper, num_eigen);
        if(!check) printf( " Problems with QR \n");

        printf("This is the T matrix \n");
        for( i = 0; i < num_eigen; ++i )
        {
            printf( "%3d  ", i);
            for( j = 0; j < i-1; ++j )
            {
                 printf( "%14.6e ",0.0);
	    }
	    if(i) printf( "%14.6e ",*(T_upper+ i-1));
            printf( "%14.6e ",1.0/(*(T_diag+i)));
	    if(i < num_eigen - 1)
            {
            	printf( "%14.6e ",*(T_upper+i));
	    }
            for( j = i+1; j < num_eigen-1; ++j )
            {
                 printf( "%14.6e ",0.0);
	    }
            printf( "\n");
        }
        printf( "\n");

	for( i = 0; i < num_eigen; ++i )
	{
		eigen[i].val = 1.0/(*(T_diag+i));
	}

	qsort( eigen, num_eigen, sizeof(EIGEN), compare_eigen);

        printf( "\n");
	printf("These are the eigenvectors of T\n");
	for( i = 0; i < num_eigen; ++i )
	{
	    for( j = 0; j < num_eigen; ++j )
	    {
		printf( "%14.6e ",*(T_evector + num_eigen*i + j));
	    }
            printf( "\n");
	}   
        printf( "\n");

        printf( "\n");
	printf("These are the lanczos vectors\n");

/* I'm actually writing out the transpose of how I'm storing it */

	for( i = 0; i < n; ++i )
	{
	    for( j = 0; j < num_eigen; ++j )
	    {
		printf( "%14.6e ",*(q + i + n*j));
	    }
            printf( "\n");
	}   
        printf( "\n");

	check = matXT(ritz, q, T_evector, n, num_eigen, num_eigen);

/* Redorder the eigenvecotors */

	for( i = 0; i < n; ++i )
	{
	    for( j = 0; j < num_eigen; ++j )
	    {
		*(Mq + num_eigen*i + j) = *(ritz + num_eigen*i + eigen[j].index);
	    }
	    printf( "\n");
	}
	printf( "\n");
	for( i = 0; i < n*num_eigen; ++i )
	{
		 *(ritz + i) = *(Mq + i);
	}

        printf( "\n");
	printf("These are the eigenvectors of the original problem\n");
	for( i = 0; i < n; ++i )
	{
	    for( j = 0; j < num_eigen; ++j )
	    {
		printf( "%14.6e ",*(ritz + num_eigen*i + j));
	    }
            printf( "\n");
	}   
        printf( "\n");

	printf("These are the eigenvalues in order\n");
	for( i = 0; i < num_eigen; ++i )
	{
		printf( "%14.6e ",eigen[i].val);
	}
	printf( "\n\n");
	for( i = 0; i < num_eigen; ++i )
	{
		printf( "%3d ",eigen[i].index);
	}
	printf( "\n\n");

	printf("These are the eigenvectors in order\n");
	for( i = 0; i < n; ++i )
	{
	    for( j = 0; j < num_eigen; ++j )
	    {
		/*printf( "%14.6e ",*(ritz + num_eigen*i + eigen[j].index));*/
		printf( "%14.6e ",*(ritz + num_eigen*i + j));
	    }
	    printf( "\n");
	}
	printf( "\n");

	return 1;
}

